//
// Copyright (c) 2015-2020 CNRS INRIA
//

#ifndef __pinocchio_parsers_urdf_utils_hpp__
#define __pinocchio_parsers_urdf_utils_hpp__

#warning "This file is now deprecated and will be removed in future releases. There is no replacement."

#endif // __pinocchio_parsers_urdf_utils_hpp__
